import shutil
from pathlib import Path
from uuid import uuid4, UUID
from typing import Union
from .metadata import Metadata, Properties


class Node:
    def __init__(self,
                 node_name: str,
                 node_type: str,
                 parent=None,
                 uuid: Union[None, UUID] = None,
                 deleted: Union[None, bool] = None,
                 metadata: Union[None, dict, Metadata] = None,
                 properties: Union[None, dict, Properties] = None,
                 protected_parent_methods: Union[None, dict] = None):
        self.__node_name__ = node_name
        self.__node_type__ = node_type
        self.__parent__ = parent
        self.__uuid__ = str(uuid) if uuid else str(uuid4())
        self.__protected_parent_methods__ = protected_parent_methods
        self.init_storage()
        if isinstance(metadata, Metadata):
            self.metadata = metadata
        elif isinstance(metadata, dict):
            self.metadata = Metadata(self, metadata)
        else:
            self.metadata = Metadata(self)
        if isinstance(properties, Properties):
            self.properties = properties
        elif isinstance(properties, dict):
            self.properties = Properties(self, properties)
        else:
            self.properties = Properties(self)
        self.properties['node_name'] = node_name
        self.properties['node_type'] = node_type
        if 'uuid' in self.properties:
            self.__uuid__ = self.properties['uuid']
        else:
            self.properties['uuid'] = self.__uuid__
        if deleted is not None:
            self.properties['deleted'] = deleted
        elif 'deleted' not in self.properties:
            self.properties['deleted'] = False

    def set_metadata(self, metadata: Union[None, dict], merge: bool = True):
        if metadata is None:
            return
        if merge:
            metadata = {**self.metadata, **metadata}
        self.metadata = Metadata(self, metadata)

    def set_properties(self, properties: Union[None, dict], merge: bool = True):
        if properties is None:
            return
        if merge:
            properties = {**self.properties, **properties}
        self.properties = Properties(self, properties)

    @property
    def uuid(self) -> str:
        return self.__uuid__

    @property
    def name(self) -> str:
        return self.__node_name__

    def rename(self, new_name: str):
        self.move_storage(self.__parent__.path / new_name)
        self.__node_name__ = new_name
        self.properties['node_name'] = new_name

    @property
    def node_type(self) -> str:
        return self.__node_type__

    def is_type(self, type_to_compare: str) -> bool:
        return self.__node_type__ == type_to_compare

    @property
    def parent(self):
        return self.__parent__

    def move_to(self, new_parent):
        if new_parent.path != self.__parent__.path:
            old_path = self.path
            self.copy_storage(new_parent.path / self.name)
            self.__parent__ = new_parent
            shutil.rmtree(old_path)
        else:
            raise FileExistsError

    def copy_to(self, new_parent):
        if new_parent.path != self.__parent__.path:
            self.copy_storage(new_parent.path / self.name)
            self.__parent__ = new_parent
        else:
            raise FileExistsError

    @property
    def path(self) -> Path:
        return self.__parent__.path / self.__node_name__

    @property
    def node_id(self) -> str:
        return self.__uuid__

    @property
    def deleted(self) -> bool:
        return self.properties['deleted']

    def delete(self):
        self.properties['deleted'] = True

    def restore(self):
        self.properties['deleted'] = False

    def init_storage(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def move_storage(self, new_path: Union[str, Path]):
        shutil.move(str(self.path), str(new_path))

    def copy_storage(self, new_path: Union[str, Path]):
        shutil.copytree(str(self.path), str(new_path), dirs_exist_ok=True)

    def purge_storage(self, confirm: bool = True, feedback: bool = False):
        if confirm and not feedback:
            return
        self.delete()
        shutil.rmtree(str(self.path))

    def __eq__(self, other: 'Node'):
        return self.__uuid__ == other.uuid

    def __hash__(self) -> int:
        return UUID(self.__uuid__).int

    def __repr__(self) -> str:
        return f"{self.__node_type__}('{self.__node_name__}')"


class Root:
    def __init__(self, root_path: Union[str, Path]):
        if isinstance(root_path, Path):
            self.__root_path__ = root_path
        else:
            self.__root_path__ = Path(root_path)

    @property
    def path(self) -> Path:
        return self.__root_path__

    def init_storage(self):
        self.__root_path__.mkdir(parents=True)



from scidb.core import DataSet, Metadata, Properties
from typing import Union, List, Callable, Any


def empty_dicts(key: str, value: Any, current_path: List[str]) -> bool:
    if isinstance(value, dict) or isinstance(value, Metadata) or isinstance(value, Properties):
        return bool(value)


def iter_node_dict(node_dict: Union[Metadata, Properties, dict],
                   path: List[str],
                   judge: Callable = empty_dicts,
                   confirm: bool = True,
                   feedback: bool = False,
                   **kwargs):
    keys_to_delete = []
    for key, value in node_dict.items():
        current_path = path + [key]
        if judge(key, value, current_path, **kwargs):
            if confirm and not feedback:
                continue
            keys_to_delete.append(key)
        if isinstance(value, dict) or isinstance(value, Metadata) or isinstance(value, Properties):
            iter_node_dict(
                node_dict=value,
                path=current_path,
                judge=judge,
                confirm=confirm,
                feedback=feedback,
                **kwargs
            )
    for key in keys_to_delete:
        node_dict.pop(key)


def clean_useless_properties(data_set: DataSet,
                             judge: Callable = empty_dicts,
                             confirm: bool = True,
                             feedback: bool = False):
    iter_node_dict(data_set.properties, [str(data_set.path), 'properties'], judge, confirm, feedback)


def clean_useless_metadata(data_set: DataSet,
                           judge: Callable = empty_dicts,
                           confirm: bool = True,
                           feedback: bool = False):
    iter_node_dict(data_set.metadata, [str(data_set.path), 'metadata'], judge, confirm, feedback)



from scidb.core import Database, Bucket, DataSet
from scidb.utils.migrator import migrate
from typing import Set, Union


class SandboxManager:
    def __init__(self, db: Database):
        self.__db__ = db

    @property
    def db(self) -> Database:
        return self.__db__

    @classmethod
    def __convert_name__(cls,
                         sandbox_name: Union[None, str] = None,
                         pure_name: Union[None, str] = None) -> str:
        if sandbox_name is not None:
            return sandbox_name.replace('__SANDBOX_', '').replace('__', '')
        elif pure_name is not None:
            return f'__SANDBOX_{pure_name}__'
        else:
            raise AssertionError

    @classmethod
    def is_sandbox(cls, bucket: Bucket) -> bool:
        return bucket.name.startswith('__SANDBOX_') and bucket.name.endswith('__')

    def create_sandbox(self, name: str) -> Bucket:
        sandbox_name = self.__convert_name__(pure_name=name)
        return self.__db__.add_bucket(sandbox_name)

    def get_sandbox(self, name: str, include_deleted: bool = False) -> Union[Bucket, None]:
        sandbox_name = self.__convert_name__(pure_name=name)
        sandbox = self.__db__.get_bucket(sandbox_name, include_deleted)
        return sandbox

    @property
    def sandboxes(self) -> Set[Bucket]:
        return set(filter(lambda bucket: self.is_sandbox(bucket), self.__db__.all_buckets))

    def list_sandbox(self) -> Set[Bucket]:
        return self.sandboxes

    def clear_sandbox(self, name: str, confirm: bool = True, feedback: bool = False):
        if confirm and not feedback:
            return
        self.delete_sandbox(name, confirm, feedback)
        self.create_sandbox(name)

    def delete_sandbox(self, name: str, confirm: bool = True, feedback: bool = False):
        if confirm and not feedback:
            return
        sandbox = self.get_sandbox(name, include_deleted=True)
        if sandbox is None:
            raise FileNotFoundError
        else:
            sandbox.delete()
            self.__db__.clear_trash(confirm, feedback)

    def delete_all_sandboxes(self, confirm: bool = True, feedback: bool = False):
        if confirm and not feedback:
            return
        for sandbox in self.sandboxes:
            sandbox.delete()
            sandbox.clear_trash(confirm, feedback)

    def migrate_sandbox(self,
                        name: str,
                        destination: Union[Bucket, DataSet],
                        delete_source: bool = True,
                        allow_overwrite: bool = False,
                        confirm: bool = True,
                        feedback: bool = False,
                        verbose: bool = True):
        if confirm and not feedback:
            return
        sandbox = self.get_sandbox(name, include_deleted=True)
        if sandbox is None:
            raise FileNotFoundError
        for data_set in sandbox.all_data_sets:
            migrate(data_set, destination, delete_source, allow_overwrite, confirm, feedback, verbose)



from ..base.backend import BackupBackend
from ..base.backup_profile import BackupProfile
from scidb.core import Database, Data
from scidb.utils.extractor import db_to_json, recover_db, get_data_list
from scidb.utils.iteration import iter_data
from typing import Tuple, List, Union, Callable
from pathlib import Path
from datetime import datetime
import json
import shutil


class LocalBackupProfile(BackupProfile):
    def __init__(self,
                 db_name: str,
                 profile_name: Union[None, str] = None,
                 time: Union[None, datetime] = None,
                 path: Union[None, str, Path] = None):
        self.__root_path__ = None
        self.set_path(path)
        super().__init__(db_name, profile_name, time)

    @property
    def path(self) -> Path:
        return self.__root_path__

    @property
    def db_json(self) -> Path:
        return self.__root_path__ / self.name

    @property
    def obj_path(self) -> Path:
        return self.__root_path__ / 'objects'

    def set_path(self, path: Union[None, str, Path]):
        if path is None or isinstance(path, Path):
            self.__root_path__ = path
        else:
            self.__root_path__ = Path(path)

    def __str__(self) -> str:
        return f"LocalBackupProfile(db='{self.__db_name__}', time='{self.time}', path='{self.__root_path__}')"

    def __repr__(self) -> str:
        return f"LocalBackupProfile(db='{self.__db_name__}', time='{self.time}', path='{self.__root_path__}')"


class LocalBackend(BackupBackend):
    def __init__(self, db_name: str, db_path: Union[str, Path], backup_path: Union[str, Path]):
        self.__db_name__ = db_name
        self.__db_path__ = db_path if isinstance(db_path, Path) else Path(db_path)
        self.__backup_path__ = backup_path if isinstance(backup_path, Path) else Path(backup_path)
        self.__db__ = Database(db_name, str(db_path))
        super().__init__()

    def ping(self) -> Union[bool, Tuple[bool, float]]:
        return True

    def connect(self):
        self.__is_connected__ = True

    def create_backup(self, verbose: bool = True, require_hash_update: bool = False) -> LocalBackupProfile:
        profile = LocalBackupProfile(db_name=self.__db_name__, time=datetime.now(), path=self.__backup_path__)
        profile.obj_path.mkdir(parents=True, exist_ok=True)
        with open(str(profile.db_json), 'w') as fp:
            json.dump(
                obj=db_to_json(self.__db_name__, self.__db_path__, verbose=verbose, require_hash_update=require_hash_update),
                fp=fp,
                indent=2
            )

        def copy_data_objs(data: Data):
            if verbose:
                print('Added:', data.name, data.path)
            dst_path = profile.obj_path / data.sha1(require_update=require_hash_update)
            if not dst_path.exists():
                shutil.copyfile(
                    src=str(data.path),
                    dst=str(dst_path)
                )

        for bucket in self.__db__.all_buckets:
            iter_data(bucket, copy_data_objs, include_deleted=True)

        return profile

    def sync_backup(self):
        pass

    def list_backups(self, db_name: Union[None, str] = None) -> List[LocalBackupProfile]:
        if db_name is None:
            db_name = self.__db_name__
        backups = [
            child
            for child in
            self.__backup_path__.glob(f'db_backup_{db_name}_*.json')
            if child.is_file()
        ]
        return [
            LocalBackupProfile(db_name=db_name, profile_name=backup.name, path=self.__backup_path__)
            for backup in backups
        ]

    def fetch_backup(self, time: datetime, db_name: Union[None, str] = None) -> Union[None, LocalBackupProfile]:
        if db_name is None:
            db_name = self.__db_name__
        profile = LocalBackupProfile(db_name=db_name, time=time, path=self.__backup_path__)
        if profile.db_json.exists():
            return profile
        else:
            return None

    def recover_from_backup(self, profile: LocalBackupProfile, new_path: Union[str, Path]):
        if isinstance(new_path, str):
            new_path = Path(new_path)
        if new_path.exists():
            raise FileExistsError

        def get_file(sha1: str) -> str:
            return str(profile.obj_path / sha1)

        with open(str(profile.db_json)) as fp:
            db_json = json.load(fp)
            recover_db(db_json, new_path, get_file=get_file)

    def clean_objects(self, confirm: bool = True, feedback: Union[Callable, bool] = False, verbose: bool = True):
        backups = [
            child
            for child in
            self.__backup_path__.glob(f'db_backup_*_*.json')
            if child.is_file()
        ]
        obj_list = set()
        for backup in backups:
            with open(str(backup), 'r') as fp:
                db_json = json.load(fp)
                obj_list = obj_list.union(get_data_list(db_json))
        obj_path = self.__backup_path__ / 'objects'
        objs = [
            child
            for child in
            obj_path.glob('*')
            if child.is_file()
        ]
        useless_objs = [obj for obj in objs if obj.name not in obj_list]
        total = len(useless_objs)
        for i, obj in enumerate(useless_objs):
            if verbose:
                print(f'[{i + 1}/{total}] Remove: {obj.name}')
            if callable(feedback):
                r = feedback(obj, i, total)
            else:
                r = feedback
            if confirm and not r:
                continue
            obj.unlink()



from .low.node import Root
from .bucket import Bucket
from typing import Set
from time import process_time


class Database(Root):
    def __init__(self, name: str, path: str, version: str = 'alpha1', verbose: bool = True):
        if verbose:
            print(f"[SciDB {version}] Loading database '{name}' @ {path} ...")
        start = process_time()
        self.__db_name__ = name
        self.__db_version__ = version
        self.__buckets__ = set()
        self.__bucket_count__ = 0
        self.__data_set_count__ = 0
        self.__data_count__ = 0
        self.__protected_methods__ = {
            'increase_bucket_count': self.__increase_bucket_count__,
            'increase_data_set_count': self.__increase_data_set_count__,
            'increase_data_count': self.__increase_data_count__
        }
        super().__init__(path)
        self.init_storage()
        self.init_buckets()
        end = process_time()
        if verbose:
            print(f"[SciDB {version}] Loaded database '{name}' "
                  f"with {self.__bucket_count__} buckets, "
                  f"{self.__data_set_count__} data sets "
                  f"and {self.__data_count__} data files "
                  f"in {end - start: .2f} seconds.")

    @property
    def bucket_count(self):
        return self.__bucket_count__

    def __increase_bucket_count__(self):
        self.__bucket_count__ += 1

    @property
    def data_set_count(self):
        return self.__data_set_count__

    def __increase_data_set_count__(self):
        self.__data_set_count__ += 1

    @property
    def data_count(self):
        return self.__data_count__

    def __increase_data_count__(self):
        self.__data_count__ += 1

    def init_buckets(self):
        children = filter(lambda child: child.is_dir(), self.path.iterdir())
        for bucket in children:
            self.__buckets__.add(Bucket(bucket.name, self, protected_parent_methods=self.__protected_methods__))

    @property
    def name(self) -> str:
        return self.__db_name__

    def rename(self, new_name: str):
        self.__db_name__ = new_name

    @property
    def version(self) -> str:
        return self.__db_version__

    @property
    def buckets(self) -> Set[Bucket]:
        return set(filter(lambda bucket: not bucket.deleted, self.__buckets__))

    @property
    def trash(self) -> Set[Bucket]:
        return set(filter(lambda bucket: bucket.deleted, self.__buckets__))

    @property
    def all_buckets(self) -> Set[Bucket]:
        return self.__buckets__

    def add_bucket(self, name: str) -> Bucket:
        if self.get_bucket(name, include_deleted=True) is not None:
            raise FileExistsError
        new_bucket = Bucket(bucket_name=name, parent=self, protected_parent_methods=self.__protected_methods__)
        self.__buckets__.add(new_bucket)
        return new_bucket

    def insert_bucket(self, bucket: Bucket) -> Bucket:
        if bucket in self.__buckets__ \
                or self.get_bucket(bucket.name, include_deleted=True) is not None \
                or self.get_bucket(bucket.uuid, include_deleted=True):
            raise FileExistsError
        if bucket.parent is not self:
            raise AssertionError
        self.__buckets__.add(bucket)
        return bucket

    def get_bucket(self, name_or_uuid: str, include_deleted: bool = False) -> Bucket:
        target = None
        search_list = self.all_buckets if include_deleted else self.buckets
        for bucket in search_list:
            if bucket.name == name_or_uuid or bucket.uuid == name_or_uuid:
                target = bucket
        return target

    def touch_bucket(self, name: str) -> Bucket:
        target = self.get_bucket(name, include_deleted=True)
        if target is None:
            target = self.add_bucket(name)
        if target.deleted:
            raise AssertionError
        return target

    def clear_trash(self, confirm: bool = True, feedback: bool = False):
        if confirm and not feedback:
            return
        for bucket in self.trash:
            bucket.purge_storage(confirm, feedback)
        for bucket in self.buckets:
            bucket.clear_trash(confirm, feedback)
        self.__buckets__ = set(self.buckets)

    def init_storage(self):
        if not self.path.exists():
            super().init_storage()



import json
import yaml
import enum
from typing import Callable, Any, List, Union
from collections.abc import Mapping


class MetadataFileType(enum.Enum):
    Json = 0
    Yaml = 1


class ObservableDict(dict):
    def __init__(self, init_dict: dict, callback: Callable):
        self.__callback__ = callback
        for item, value in init_dict.items():
            if isinstance(value, dict):
                init_dict[item] = ObservableDict(value, callback)
        super().__init__(init_dict)

    def __setitem__(self, item: str, value: Any):
        value = ObservableDict(value, self.__callback__) if isinstance(value, dict) else value
        super().__setitem__(item, value)
        self.__callback__(item, value)

    def __getitem__(self, item: str) -> Any:
        return super().__getitem__(item)

    def pop(self, key: str) -> Any:
        super().pop(key, None)
        self.__callback__()

    def set(self, item: str, value: Any):
        self.__setitem__(item, value)

    def get(self, item: str) -> Any:
        return self.__getitem__(item)

    def set_by_path(self, path: List[str], value: Any):
        data = self
        length = len(path)
        for i, key in enumerate(path):
            if i < length - 1:
                if key not in data:
                    data[key] = dict()
                data = data[key]
            else:
                data[key] = value

    def get_by_path(self, path: List[str]) -> Any:
        data = self
        for key in path:
            if key not in data:
                return None
            data = data[key]
        return data

    def delete_by_path(self, path: List[str]):
        data = self
        length = len(path)
        for i, key in enumerate(path):
            if i < length - 1:
                if key not in data:
                    return
                data = data[key]
            else:
                data.pop(key)

    def to_dict(self) -> dict:
        return {key: dict(value) if isinstance(value, ObservableDict) else value for key, value in self.items()}


class NodeDict(Mapping):
    def __init__(self,
                 node,
                 filename: str,
                 data: Union[None, dict] = None,
                 file_type: MetadataFileType = MetadataFileType.Json):
        self.__node__ = node
        self.__filename__ = filename
        self.__data__ = ObservableDict({}, self.callback)
        self.__file_type__ = file_type
        if data is not None:
            self.__data__ = data if isinstance(data, ObservableDict) else ObservableDict(data, self.callback)
            self.callback()
        else:
            self.__load_data__()

    def __load_data__(self):
        full_path = self.__node__.path / self.__filename__
        if not full_path.exists() or not full_path.is_file():
            return
        else:
            full_path = str(full_path)
            with open(full_path) as file:
                if self.__file_type__ == MetadataFileType.Yaml:
                    self.__data__ = ObservableDict(yaml.load(file), self.callback)
                elif self.__file_type__ == MetadataFileType.Json:
                    self.__data__ = ObservableDict(json.load(file), self.callback)
                else:
                    raise NotImplementedError('The type is not supported.')

    def callback(self, item: str = None, value: Any = None):
        full_path = str(self.__node__.path / self.__filename__)
        with open(full_path, mode='w') as file:
            if self.__file_type__ == MetadataFileType.Yaml:
                yaml.dump(self.__data__.to_dict(), file)
            elif self.__file_type__ == MetadataFileType.Json:
                json.dump(self.__data__, file)
            else:
                raise NotImplementedError('The type is not supported.')

    def __setitem__(self, item: str, value: Any):
        self.__data__[item] = value

    def __getitem__(self, item: str) -> Any:
        return self.__data__[item]

    def __contains__(self, item: str):
        return item in self.__data__

    def set(self, item: str, value: Any):
        self.__setitem__(item, value)

    def get(self, item: str) -> Any:
        return self.__getitem__(item)

    def set_by_path(self, path: List[str], value: Any):
        self.__data__.set_by_path(path, value)

    def get_by_path(self, path: List[str]) -> Any:
        return self.__data__.get_by_path(path)

    def delete_by_path(self, path: List[str]):
        self.__data__.delete_by_path(path)

    @property
    def data(self) -> ObservableDict:
        return self.__data__

    def __str__(self) -> str:
        return str(self.__data__.to_dict())

    def __repr__(self) -> str:
        return str(self.__data__.to_dict())

    def __iter__(self):
        return iter(self.__data__)

    def __len__(self):
        return len(self.__data__)

    def pop(self, key: str):
        self.__data__.pop(key)


class Metadata(NodeDict):
    def __init__(self,
                 node,
                 data: Union[None, dict] = None,
                 file_type: MetadataFileType = MetadataFileType.Json):
        if file_type == MetadataFileType.Yaml:
            filename = 'metadata.yml'
        elif file_type == MetadataFileType.Json:
            filename = 'metadata.json'
        else:
            raise NotImplementedError('The type is not supported.')
        super().__init__(node, data=data, filename=filename, file_type=file_type)


class Properties(NodeDict):
    def __init__(self,
                 node,
                 data: Union[None, dict] = None,
                 file_type: MetadataFileType = MetadataFileType.Json):
        if file_type == MetadataFileType.Yaml:
            filename = 'properties.yml'
        elif file_type == MetadataFileType.Json:
            filename = 'properties.json'
        else:
            raise NotImplementedError('The type is not supported.')
        super().__init__(node, data=data, filename=filename, file_type=file_type)



"""
Command list:

bucket select [<name> | <uuid>]
bucket create <name>
bucket list [all | deleted]
bucket rm [<name> | <uuid>]
bucket restore [<name> | <uuid> | all]
bucket clean [-f] [<name> | <uuid> | all]
"""

from scidb.core import Database, Bucket
from typing import List, Set, Union
import scidb.client.global_env as global_env


usage = """\
 1 | > bucket select [<name> | <uuid>]
   |   To select a bucket.
 2 | > bucket create <name>
   |   To create a bucket with given name.
 3 | > bucket list [all | deleted]
   |   To list (all / deleted) buckets.
 4 | > bucket rm [<name> | <uuid>]
   |   To delete a bucket with given name or uuid.
 5 | > bucket restore [<name> | <uuid> | all]
   |   To restore a deleted bucket.
 6 | > bucket clean [-f] [<name> | <uuid> | all]
   |   To delete trash of a bucket / all buckets.
"""

select_usage = """\
> bucket select [<name> | <uuid>]

To select a bucket.
<name> | OPTIONAL | name of the bucket,
<uuid> | OPTIONAL | uuid of the bucket.
"""

create_usage = """\
> bucket create <name>

To create a bucket with given name.
<name> | REQUIRED | name of the new bucket.
"""

list_usage = """\
> bucket list [all | deleted]

To list (all / deleted) buckets.
all     | OPTIONAL | list all buckets.
deleted | OPTIONAL | list deleted buckets.
"""

rm_usage = """\
> bucket rm [<name> | <uuid>]

To delete a bucket with given name or uuid.
<name> | OPTIONAL | name of the bucket,
<uuid> | OPTIONAL | uuid of the bucket.
"""

restore_usage = """\
> bucket restore [<name> | <uuid> | all]

To restore a deleted bucket.
<name> | OPTIONAL | name of the bucket,
<uuid> | OPTIONAL | uuid of the bucket,
all    | OPTIONAL | restore all buckets.
"""

clean_usage = """\
> bucket clean [-f] [<name> | <uuid> | all]

To delete trash of a bucket / all buckets.
-f     | OPTIONAL | force clean (without confirmation),
<name> | OPTIONAL | name of the bucket,
<uuid> | OPTIONAL | uuid of the bucket,
all    | OPTIONAL | clean all buckets.
"""


def handler(args: List[str]):
    if len(args) < 1:
        print(usage)
        return
    if global_env.CONNECTED_DATABASE is None:
        print('No database connected.')
        return
    if not isinstance(global_env.CONNECTED_DATABASE, Database):
        print('Internal error.')
        exit(-1)
        return
    if args[0] == 'select':
        if len(args) != 2:
            print(select_usage)
            return
        select_bucket(args[1])
    elif args[0] == 'create':
        if len(args) != 2:
            print(create_usage)
            return
        create_bucket(args[1])
    elif args[0] == 'list':
        if len(args) == 1:
            list_bucket()
        elif len(args) == 2 and args[1] in ['all', 'deleted']:
            list_bucket(args[1])
        else:
            print(list_usage)
            return
    elif args[0] == 'rm':
        if len(args) != 2:
            print(rm_usage)
            return
        rm_bucket(args[1])
    elif args[0] == 'restore':
        if len(args) != 2:
            print(restore_usage)
            return
        if 'all' in args:
            restore_bucket()
        else:
            restore_bucket(args[1])
    elif args[0] == 'clean':
        if len(args) not in [2, 3]:
            print(clean_usage)
            return
        if '-f' in args:
            args.remove('-f')
            if args[1] == 'all':
                clean_bucket(None, confirm=True, feedback=True)
            else:
                clean_bucket(args[1], confirm=True, feedback=True)
        else:
            if args[1] == 'all':
                clean_bucket(None, confirm=True,
                             feedback=input('Would you like to clean all buckets? Y/[N]') == 'Y')
            else:
                clean_bucket(args[1], confirm=True,
                             feedback=input('Would you like to clean this bucket? Y/[N]') == 'Y')
    else:
        print(usage)
        return


def select_bucket(name_or_uuid: str):
    global_env.SELECTED_BUCKET = global_env.CONNECTED_DATABASE.get_bucket(name_or_uuid, include_deleted=True)


def create_bucket(bucket_name: str):
    global_env.SELECTED_BUCKET = global_env.CONNECTED_DATABASE.add_bucket(bucket_name)


def print_buckets(buckets: Set[Bucket]):
    print('  No. |   bucket name   |                 uuid                ')
    print('--------------------------------------------------------------')
    for i, bucket in enumerate(buckets):
        print(f'{i: 5} | {bucket.name: <15} | {bucket.uuid}')
    if len(buckets) == 0:
        print('                         No records.                          ')
    print('--------------------------------------------------------------')


def list_bucket(bucket_filter: Union[None, str] = None):
    if bucket_filter is None:
        print_buckets(global_env.CONNECTED_DATABASE.buckets)
    elif bucket_filter == 'all':
        print_buckets(global_env.CONNECTED_DATABASE.all_buckets)
    elif bucket_filter == 'deleted':
        print_buckets(global_env.CONNECTED_DATABASE.trash)
    else:
        return


def rm_bucket(name_or_uuid: str):
    bucket = global_env.CONNECTED_DATABASE.get_bucket(name_or_uuid)
    if bucket is not None:
        bucket.delete()


def restore_bucket(name_or_uuid: Union[str, None] = None):
    if name_or_uuid is None:
        for deleted in global_env.CONNECTED_DATABASE.trash:
            deleted.restore()
    else:
        target = global_env.CONNECTED_DATABASE.get_bucket(name_or_uuid, include_deleted=True)
        if target is not None:
            target.restore()
        else:
            print('No such bucket.')


def clean_bucket(name_or_uuid: Union[str, None], confirm: bool = True, feedback: bool = False):
    if confirm and not feedback:
        print('User cancelled.')
        return
    if name_or_uuid is None:
        global_env.CONNECTED_DATABASE.clear_trash(confirm, feedback)
    else:
        bucket = global_env.CONNECTED_DATABASE.get_bucket(name_or_uuid)
        if bucket is None:
            print('No such bucket.')
        else:
            bucket.clear_trash(confirm, feedback)



from scidb.core import Database, Bucket, DataSet, Data
from typing import Union, Callable
from json import loads
from pathlib import Path
from tempfile import TemporaryDirectory


def db_to_json(db_name: str, db_path: str, verbose: bool = True, require_hash_update: bool = False):
    db = Database(db_name, db_path)
    if verbose:
        print('Extract Database:', db_name, '@', db_path)
    results = dict()
    results['properties'] = dict()
    results['properties']['name'] = db.name
    results['properties']['path'] = str(db.path)
    results['properties']['version'] = db.version
    results['buckets'] = dict()
    for bucket in db.all_buckets:
        results['buckets'][bucket.name] = dict()
        results['buckets'][bucket.name]['properties'] = bucket.properties.data.to_dict()
        results['buckets'][bucket.name]['metadata'] = bucket.metadata.data.to_dict()
        results['buckets'][bucket.name]['children'] = dict()
        bucket_to_json(bucket,
                       results=results['buckets'][bucket.name]['children'],
                       verbose=verbose,
                       require_hash_update=require_hash_update)
    return results


def bucket_to_json(bucket: Bucket, results: dict, verbose: bool = True, require_hash_update: bool = False):
    if verbose:
        print('Extract Bucket:', bucket.name)
    for data_set in bucket.all_data_sets:
        data_set_to_json(data_set, results, verbose=verbose, require_hash_update=require_hash_update)


def data_set_to_json(data_set: DataSet, results: dict, verbose: bool = True, require_hash_update: bool = False):
    if verbose:
        print('Extract DataSet:', data_set.name)
    results[data_set.name] = dict()
    results[data_set.name]['properties'] = data_set.properties.data.to_dict()
    results[data_set.name]['metadata'] = data_set.metadata.data.to_dict()
    results[data_set.name]['children'] = dict()
    results[data_set.name]['data'] = dict()
    for child in data_set.all_data_sets:
        data_set_to_json(child,
                         results[data_set.name]['children'],
                         verbose=verbose,
                         require_hash_update=require_hash_update)
    for data in data_set.data:
        data_to_json(data, results[data_set.name]['data'], verbose=verbose, require_hash_update=require_hash_update)


def data_to_json(data: Data, results: dict, verbose: bool = True, require_hash_update: bool = False):
    if verbose:
        print('Extract Data:', data.name)
    results[data.name] = data.sha1(require_update=require_hash_update)


def recover_db(db_json: Union[dict, str],
               new_path: Union[str, Path],
               get_file: Union[None, Callable] = None,
               verbose: bool = True,
               overwrite: bool = False):
    if verbose:
        print('Recover Database to', new_path)
    if isinstance(db_json, str):
        db_json = loads(db_json)
    if isinstance(new_path, str):
        new_path = Path(new_path)
    if new_path.exists() and not overwrite:
        raise FileExistsError
    assert 'properties' in db_json
    properties = db_json['properties']
    assert 'name' in properties and 'buckets' in db_json
    name = properties['name']
    version = properties.get('version', 'alpha1')
    buckets = db_json['buckets']
    assert isinstance(buckets, dict)
    db = Database(name, path=str(new_path), version=version)
    recover_buckets(db, buckets, get_file=get_file, verbose=verbose)


def recover_buckets(db: Database,
                    buckets: dict,
                    get_file: Union[None, Callable] = None,
                    verbose: bool = True):
    for bucket_name, bucket_info in buckets.items():
        if verbose:
            print('Recover Bucket', bucket_name)
        properties = bucket_info['properties']
        metadata = bucket_info['metadata']
        children = bucket_info['children']
        assert isinstance(children, dict)
        new_bucket = Bucket(
            bucket_name=bucket_name,
            parent=db,
            metadata=metadata,
            properties=properties
        )
        new_bucket = db.insert_bucket(new_bucket)
        recover_data_sets(new_bucket, children, get_file=get_file, verbose=verbose)


def recover_data_sets(parent: Union[Bucket, DataSet],
                      data_sets: dict,
                      get_file: Union[None, Callable] = None,
                      verbose: bool = True):
    for data_set_name, data_set_info in data_sets.items():
        if verbose:
            print('Recover DataSet', data_set_name)
        properties = data_set_info['properties']
        metadata = data_set_info['metadata']
        children = data_set_info['children']
        data = data_set_info['data']
        new_data_set = DataSet(
            data_set_name=data_set_name,
            parent=parent,
            metadata=metadata,
            properties=properties
        )
        parent.insert_data_set(new_data_set)
        recover_data_sets(new_data_set, children, get_file=get_file, verbose=verbose)
        recover_data(new_data_set, data, get_file=get_file, verbose=verbose)


def recover_data(data_set: DataSet,
                 data: dict,
                 file: Union[None, str, Path] = None,
                 get_file: Union[None, Callable] = None,
                 verbose: bool = True):
    for data_name, data_sha1 in data.items():
        if verbose:
            print('Recover DataSet', data_name)
        new_data = data_set.add_data(data_name)
        if file:
            new_data.import_file(file, confirm=False)
        elif get_file is not None:
            new_data.import_file(get_file(data_sha1), confirm=False)
        else:
            print(f'WARNING: '
                  f'No file imported for data `{data_name}` with SHA1 value `{data_sha1}`. '
                  f'This may cause data loss.')


def get_data_list(db_json: dict) -> set:
    data_list = set()
    temp_dir = TemporaryDirectory()
    empty_file = Path(temp_dir.name) / 'empty.file'
    with open(empty_file, 'w') as fp:
        fp.write('')

    def list_objs(sha1: str) -> str:
        data_list.add(sha1)
        return str(empty_file)

    recover_db(db_json, temp_dir.name, list_objs, overwrite=True)
    temp_dir.cleanup()
    return data_list



from .low.node import Node
from .data_set import DataSet
from .low.metadata import Metadata, Properties
from uuid import UUID
from typing import Set, Union


class Bucket(Node):
    def __init__(self,
                 bucket_name: str,
                 parent,
                 uuid: Union[None, str, UUID] = None,
                 deleted: Union[None, bool] = None,
                 metadata: Union[None, Metadata] = None,
                 properties: Union[None, Properties] = None,
                 protected_parent_methods: Union[None, dict] = None):
        self.__data_sets__ = set()
        super().__init__(
            node_name=bucket_name,
            node_type='Bucket',
            parent=parent,
            uuid=uuid,
            deleted=deleted,
            metadata=metadata,
            properties=properties,
            protected_parent_methods=protected_parent_methods
        )
        self.__protected_parent_methods__['increase_bucket_count']()
        self.init_data_sets()

    @property
    def database(self):
        return self.parent

    @property
    def db(self):
        return self.parent

    def init_data_sets(self):
        children = filter(lambda child: child.is_dir(), self.path.iterdir())
        for data_set in children:
            self.__data_sets__.add(DataSet(data_set.name, parent=self, bucket=self,
                                           protected_parent_methods=self.__protected_parent_methods__))

    def add_data_set(self, name: str) -> DataSet:
        if self.get_data_set(name, include_deleted=True) is not None:
            raise FileExistsError
        new_data_set = DataSet(data_set_name=name, parent=self,
                               bucket=self, protected_parent_methods=self.__protected_parent_methods__)
        self.__data_sets__.add(new_data_set)
        return new_data_set

    def insert_data_set(self, data_set: DataSet) -> DataSet:
        if data_set in self.__data_sets__ \
                or self.get_data_set(data_set.name, include_deleted=True) is not None \
                or self.get_data_set(data_set.uuid, include_deleted=True) is not None:
            raise FileExistsError
        if data_set.parent is not self:
            raise AssertionError
        self.__data_sets__.add(data_set)
        return data_set

    def get_data_set(self, name_or_uuid: str, include_deleted: bool = False) -> DataSet:
        target = None
        search_list = self.all_data_sets if include_deleted else self.data_sets
        for data_set in search_list:
            if data_set.name == name_or_uuid or data_set.uuid == name_or_uuid:
                target = data_set
        return target

    def touch_data_set(self, name: str) -> DataSet:
        target = self.get_data_set(name, include_deleted=True)
        if target is None:
            target = self.add_data_set(name)
        if target.deleted:
            raise AssertionError
        return target

    @property
    def data_sets(self) -> Set[DataSet]:
        return set(filter(lambda data_set: not data_set.deleted, self.__data_sets__))

    @property
    def trash(self) -> Set[DataSet]:
        return set(filter(lambda data_set: data_set.deleted, self.__data_sets__))

    @property
    def all_data_sets(self) -> Set[DataSet]:
        return self.__data_sets__

    def clear_trash(self, confirm: bool = True, feedback: bool = False):
        if confirm and not feedback:
            return
        for data_set in self.trash:
            data_set.purge_storage(confirm, feedback)
        for data_set in self.data_sets:
            data_set.clear_trash(confirm, feedback)
        self.__data_sets__ = set(self.data_sets)



from ..base.backend import BackupBackend
from ..base.backup_profile import BackupProfile
from scidb.core import Database, Data
from scidb.utils.extractor import db_to_json, recover_db, get_data_list
from scidb.utils.iteration import iter_data
from typing import Tuple, List, Union, Callable
from pathlib import Path
from datetime import datetime
from minio import Minio
from minio.error import NoSuchKey, InvalidBucketError, NoSuchBucket
from urllib3.poolmanager import PoolManager
from tempfile import TemporaryDirectory
import json


class MinioBackupProfile(BackupProfile):
    def __init__(self,
                 db_name: str,
                 profile_name: Union[None, str] = None,
                 time: Union[None, datetime] = None,
                 obj_bucket_name: str = 'scidb-objects',
                 backup_bucket_name: str = 'scidb-backups'):
        self.obj_bucket_name = obj_bucket_name
        self.backup_bucket_name = backup_bucket_name
        self.__temp_dir__ = TemporaryDirectory()
        self.__temp_path__ = Path(self.__temp_dir__.name)
        self.obj_list = dict()
        super().__init__(db_name, profile_name, time)

    @property
    def temp_path(self) -> Path:
        return self.__temp_path__

    @property
    def db_json(self) -> Path:
        return self.__temp_path__ / self.name

    def remove_temp(self):
        self.__temp_dir__.cleanup()

    def __str__(self) -> str:
        return f"MinioBackupProfile(" \
               f"db='{self.__db_name__}', " \
               f"time='{self.time}', " \
               f"obj_bucket='{self.obj_bucket_name}', " \
               f"backup_bucket='{self.backup_bucket_name}')"

    def __repr__(self) -> str:
        return f"MinioBackupProfile(" \
               f"db='{self.__db_name__}', " \
               f"time='{self.time}', " \
               f"obj_bucket='{self.obj_bucket_name}', " \
               f"backup_bucket='{self.backup_bucket_name}')"


class MinioBackend(BackupBackend):
    def __init__(self,
                 db_name: str,
                 db_path: Union[str, Path],
                 endpoint: str,
                 access_key: str,
                 secret_key: str,
                 secure: bool = True,
                 region: Union[str, None] = None,
                 http_client: Union[PoolManager, None] = None,
                 obj_bucket_name: str = 'scidb-objects',
                 backup_bucket_name: str = 'scidb-backups'):
        self.__db_name__ = db_name
        self.__db_path__ = db_path if isinstance(db_path, Path) else Path(db_path)
        self.__endpoint__ = endpoint
        self.__access_key__ = access_key
        self.__secret_key__ = secret_key
        self.__secure__ = secure
        self.__region__ = region
        self.__http_client__ = http_client
        self.obj_bucket_name = obj_bucket_name
        self.backup_bucket_name = backup_bucket_name
        self.__server__ = Minio(
            endpoint,
            access_key,
            secret_key,
            secure=secure,
            region=region,
            http_client=http_client
        )
        self.__db__ = Database(db_name, str(db_path))
        self.__current_profile__: Union[None, MinioBackupProfile] = None
        super().__init__()

    @property
    def server(self):
        return self.__server__

    def init_remote_storage(self) -> bool:
        assert self.__current_profile__ is not None
        if not self.__server__.bucket_exists(self.__current_profile__.obj_bucket_name):
            self.__server__.make_bucket(self.__current_profile__.obj_bucket_name)
        if not self.__server__.bucket_exists(self.__current_profile__.backup_bucket_name):
            self.__server__.make_bucket(self.__current_profile__.backup_bucket_name)
        return self.__server__.bucket_exists(self.__current_profile__.obj_bucket_name) \
               and self.__server__.bucket_exists(self.__current_profile__.backup_bucket_name)

    def ping(self) -> Union[bool, Tuple[bool, float]]:
        return True

    def connect(self):
        self.__is_connected__ = True

    def exists_object(self, bucket_name: str, object_name: str) -> bool:
        try:
            self.__server__.remove_incomplete_upload(bucket_name, object_name)
            return self.__server__.stat_object(bucket_name, object_name) is not None
        except NoSuchKey:
            return False
        except InvalidBucketError:
            return False
        except NoSuchBucket:
            return False

    def create_backup(self, verbose: bool = True, require_hash_update: bool = False) -> MinioBackupProfile:
        profile = MinioBackupProfile(
            db_name=self.__db_name__,
            time=datetime.now(),
            backup_bucket_name=self.backup_bucket_name,
            obj_bucket_name=self.obj_bucket_name
        )
        with open(str(profile.db_json), 'w') as fp:
            json.dump(
                obj=db_to_json(self.__db_name__,
                               self.__db_path__,
                               verbose=verbose,
                               require_hash_update=require_hash_update),
                fp=fp,
                indent=2
            )

        if verbose:
            print('List remote objects.')
        remote_objs = [obj.object_name for obj in self.__server__.list_objects(profile.obj_bucket_name)]
        print(f'Found {len(remote_objs)} objects on remote server.')

        def list_data_objs(data: Data):
            h = data.sha1(require_update=require_hash_update)
            if h not in profile.obj_list and h not in remote_objs:
                if verbose:
                    print('Added:', data.name, data.path)
                profile.obj_list[h] = data.path

        for bucket in self.__db__.all_buckets:
            iter_data(bucket, list_data_objs, include_deleted=True)

        self.__current_profile__ = profile
        return profile

    def sync_backup(self, verbose: bool = True):
        if self.__current_profile__ is None:
            raise AssertionError('Backup has not been created.')
        self.init_remote_storage()
        self.__server__.fput_object(
            self.__current_profile__.backup_bucket_name,
            self.__current_profile__.name,
            str(self.__current_profile__.db_json)
        )
        total = len(self.__current_profile__.obj_list)
        for i, (name, path) in enumerate(self.__current_profile__.obj_list.items()):
            if verbose:
                print(f'[{i + 1}/{total}] Sync: {name} @ {path}')
            self.__server__.fput_object(
                self.__current_profile__.obj_bucket_name,
                name,
                path
            )

    def list_backups(self, db_name: Union[None, str] = None) -> List[MinioBackupProfile]:
        if db_name is None:
            db_name = self.__db_name__
        if not self.__server__.bucket_exists(self.backup_bucket_name):
            return []
        else:
            return [
                MinioBackupProfile(
                    db_name=db_name,
                    profile_name=backup.object_name,
                    obj_bucket_name=self.obj_bucket_name,
                    backup_bucket_name=self.backup_bucket_name
                )
                for backup in self.__server__.list_objects(self.backup_bucket_name)
            ]

    def fetch_backup(self, time: datetime, db_name: Union[None, str] = None) -> Union[None, MinioBackupProfile]:
        if db_name is None:
            db_name = self.__db_name__
        profile = MinioBackupProfile(
            db_name=db_name,
            time=time,
            obj_bucket_name=self.obj_bucket_name,
            backup_bucket_name=self.backup_bucket_name
        )
        if self.exists_object(self.backup_bucket_name, profile.name):
            return profile
        else:
            return None

    def recover_from_backup(self, profile: MinioBackupProfile, new_path: Union[str, Path]):
        if isinstance(new_path, str):
            new_path = Path(new_path)
        if new_path.exists():
            raise FileExistsError
        self.__server__.fget_object(
            profile.backup_bucket_name,
            profile.name,
            str(profile.db_json)
        )

        def get_file(sha1: str) -> str:
            file_path = str(profile.temp_path / sha1)
            self.__server__.fget_object(
                profile.obj_bucket_name,
                sha1,
                file_path
            )
            return file_path

        with open(str(profile.db_json)) as fp:
            db_json = json.load(fp)
            recover_db(db_json, new_path, get_file=get_file)

        profile.remove_temp()

    def clean_objects(self, confirm: bool = True, feedback: Union[Callable, bool] = False, verbose: bool = True):
        if not self.__server__.bucket_exists(self.backup_bucket_name) \
                or not self.__server__.bucket_exists(self.obj_bucket_name):
            return
        backups = self.__server__.list_objects(self.backup_bucket_name)
        obj_list = set()
        for backup in backups:
            response = self.__server__.get_object(self.backup_bucket_name, backup.object_name)
            data = response.read().decode('utf-8')
            db_json = json.loads(data)
            obj_list = obj_list.union(get_data_list(db_json))
        objs = self.__server__.list_objects(self.obj_bucket_name)
        useless_objs = [obj for obj in objs if obj.object_name not in obj_list]
        total = len(useless_objs)
        for i, obj in enumerate(useless_objs):
            if verbose:
                print(f'[{i + 1}/{total}] Remove: {obj.object_name}')
            if callable(feedback):
                r = feedback(obj, i, total)
            else:
                r = feedback
            if confirm and not r:
                continue
            self.__server__.remove_object(self.obj_bucket_name, obj.object_name)



"""
Command list:

data create <name>
data list
data rm [-f] <name>
data import <name> <src_path>
data export <name> <dst_path>
"""

from scidb.core import DataSet, Data
from typing import List, Set, Union
from pathlib import Path
import scidb.client.global_env as global_env


usage = """\
 1 | > data create <name>
   |   To create a data file.
 2 | > data list
   |   To list all data files in current directory.
 3 | > data rm [-f] <name>
   |   To delete a data file with given name.
 4 | > data import <name> <src_path>
   |   To import a data file.
 5 | > data export <name> <dst_path>
   |   To export a data file.
"""

create_usage = """\
> data create <name>

To create a data file.
<name> | REQUIRED | name of the new data file.
"""

list_usage = """\
> data list

To list all data files in current directory.
"""

rm_usage = """\
> data rm [-f] <name>

To delete a data file with given name.
-f     | OPTIONAL | force delete (without confirmation),
<name> | OPTIONAL | name of the data file.
"""

import_usage = """\
> data import [-f] <name> <src_path>

To import a data file.
-f         | OPTIONAL | force import (allow overwrite),
<name>     | REQUIRED | name of the data file,
<src_path> | REQUIRED | full path to the source file.
"""

export_usage = """\
> data export [-f] <name> <dst_path>

To export a data file.
-f         | OPTIONAL | force import (allow overwrite),
<name>     | REQUIRED | name of the data file,
<dst_path> | REQUIRED | full path to the destination file.
"""


def handler(args: List[str]):
    if len(args) < 1:
        print(usage)
        return
    if global_env.CURRENT_DATASET is None:
        print('No dataset selected.')
        return
    if not isinstance(global_env.CURRENT_DATASET, DataSet):
        print('Internal error.')
        exit(-1)
        return
    if args[0] == 'create':
        if len(args) != 2:
            print(create_usage)
            return
        create_data(args[1])
    elif args[0] == 'list':
        if len(args) != 1:
            print(list_usage)
            return
        list_data()
    elif args[0] == 'rm':
        if len(args) not in [2, 3]:
            print(rm_usage)
            return
        if '-f' in args:
            args.remove('-f')
            rm_data(args[1], confirm=True, feedback=True)
        else:
            rm_data(args[1], confirm=True,
                    feedback=input('Would you really want to PERMANENTLY delete this data? Y/[N]') == 'Y')
    elif args[0] == 'import':
        if len(args) not in [3, 4]:
            print(import_usage)
            return
        if '-f' in args:
            args.remove('-f')
            import_data(args[1], args[2], allow_overwrite=True)
        else:
            import_data(args[1], args[2])
    elif args[0] == 'export':
        if len(args) not in [3, 4]:
            print(export_usage)
            return
        if '-f' in args:
            args.remove('-f')
            export_data(args[1], args[2], allow_overwrite=True)
        else:
            export_data(args[1], args[2])
    else:
        print(usage)
        return


def get_current_dataset() -> DataSet:
    return global_env.CURRENT_DATASET


def create_data(name: str):
    global_env.CURRENT_DATASET.add_data(name)


def print_data(data: Set[Data]):
    print('  No. |    data name    ')
    print('------------------------')
    for i, d in enumerate(data):
        print(f'{i: 5} | {d.name: <15}')
    if len(data) == 0:
        print('       No records.      ')
    print('------------------------')


def list_data():
    print_data(global_env.CURRENT_DATASET.data)


def rm_data(name: str, confirm: bool = True, feedback: bool = False):
    if confirm and not feedback:
        print('User cancelled.')
        return
    global_env.CURRENT_DATASET.delete_data(name, confirm, feedback)


def import_data(name: str, src_path: Union[str, Path], allow_overwrite: bool = False):
    if global_env.CURRENT_DATASET.get_data(name) is not None and not allow_overwrite:
        print('Overwrite is not permitted. Try with `-f`.')
        return
    if global_env.CURRENT_DATASET.get_data(name) is None:
        global_env.CURRENT_DATASET.add_data(name)
    global_env.CURRENT_DATASET.get_data(name).import_file(src_path, allow_overwrite, confirm=True, feedback=True)


def export_data(name: str, dst_path: Union[str, Path], allow_overwrite: bool = False):
    if not isinstance(dst_path, Path):
        dst_path = Path(dst_path)
    if dst_path.exists() and not allow_overwrite:
        print('Overwrite is not permitted. Try with `-f`.')
        return
    data = global_env.CURRENT_DATASET.get_data(name)
    if data is not None:
        data.export_file(dst_path, allow_overwrite)
    else:
        print('No such data.')
